"""
The following exceptions may be raised during the course of using
:py:class:`tornado_aws.client.AWSClient` and
:py:class:`tornado_aws.client.AsyncAWSClient`:

"""


class AWSClientException(Exception):
    """Base exception class for AWSClient

    :ivar msg: The error message

    """
    fmt = 'An error occurred'

    def __init__(self, **kwargs):
        super(AWSClientException, self).__init__(self.fmt.format(**kwargs),
                                                 kwargs)


class CurlNotInstalledError(Exception):
    """Raised when the CurlAsyncHTTPClient is requested but pycurl is not
    installed.

    """
    fmt = 'use_curl was specified but pycurl is not installed'


class AWSError(AWSClientException):
    """Raised when the credentials could not be located."""
    fmt = '{message}'


class ConfigNotFound(AWSClientException):
    """The configuration file could not be parsed.

    :ivar path: The path to the config file

    """
    fmt = 'The config file could not be found ({path})'


class ConfigParserError(AWSClientException):
    """Error raised when parsing a configuration file with
    :py:class`configparser.RawConfigParser`

    :ivar path: The path to the config file

    """
    fmt = 'Unable to parse config file ({path})'


class LocalCredentialsError(AWSClientException):
    """Raised when the credentials could not be located."""
    fmt = 'Cant reset local credentials'


class NoCredentialsError(AWSClientException):
    """Raised when the credentials could not be located."""
    fmt = 'Credentials not found'


class NoProfileError(AWSClientException):
    """Raised when the specified profile could not be located.

    :ivar path: The path to the config file
    :ivar profile: The profile that was specified

    """
    fmt = 'Profile ({profile}) not found ({path})'


class RequestException(AWSClientException):
    """Raised when a request failed due to a network issue.

    :ivar error: The error which occured

    """
    fmt = 'An error occured making a request {error}'



"""
The :py:class:`AWSClient` and :py:class:`AsyncAWSClient` implement low-level
AWS clients. The clients provide only the mechanism for submitted signed HTTP
requests to the AWS APIs and are generally meant to be used by service specific
client API implementations.

"""
import datetime
import hashlib
import hmac
import json
import logging
import os
import socket
import sys
try:
    from urllib.parse import urlparse
except ImportError:  # pragma: nocover
    from urlparse import urlparse

try:
    from urllib.parse import quote
except ImportError:  # pragma: nocover
    from urllib import quote


from tornado import concurrent, httpclient, ioloop

try:
    from tornado import curl_httpclient
except ImportError:  # pragma: nocover
    curl_httpclient = None

from tornado_aws import config, exceptions, txml

LOGGER = logging.getLogger(__name__)

MIME_AWZ_JSON = 'application/x-amz-json-1.1'

PYTHON3 = True if sys.version_info > (3, 0, 0) else False

_AWZ_CONTENT_TYPES = [
    'application/json',
    'application/x-amz-json-1.0',
    'application/x-amz-json-1.1'
]

_REFRESH_EXCEPTIONS = [
    'AuthFailure',
    'AuthMissingFailure',
    'AWS.InvalidAccount',
    'ExpiredTokenException',
    'InvalidSignatureException',
    'MissingAuthenticationTokenException',
    'UnrecognizedClientException'
]

_REFRESH_XML_EXCEPTIONS = [
    'AuthFailure',
    'ExpiredToken',
    'InvalidClientTokenId',
    'InvalidSecurity',
    'MissingAuthenticationToken',
    'SignatureDoesNotMatch'
]


_HEADER_FORMAT = '{0} Credential={1}/{2}, SignedHeaders={3}, Signature={4}'


class AWSClient(object):
    """Implement a low level AWS client that performs the request signing
    required for AWS API requests.

    ``AWSClient`` uses the same configuration method and environment
    variables as the AWS CLI. For configuration information visit the "Getting
    Set Up" section of the `AWS Command Line Interface user guide
    <http://docs.aws.amazon.com/cli/latest/userguide/>`_.

    When creating the ``AWSClient`` instance you need to specify the
    ``service`` that you will be interacting with. This value is used when
    signing the request headers and must match the service values as specified
    in the `AWS General Reference documentation
    <http://docs.aws.amazon.com/general/latest/gr/Welcome.html>`_.

    The AWS configuration profile can be set when creating the
    ``AWSClient`` instance or by setting the ``AWS_DEFAULT_PROFILE``
    environment variable. If neither are set, ``default`` will be used.

    The AWS region is set by reading in configuration or by the
    ``AWS_DEFAULT_REGION`` environment variable. If neither or set, it will
    attempt to be set by invoking the EC2 Instance Metadata and user data API,
    if available.

    The AWS access key can be set when creating a new instance. If it's not
    passed in when creating the ``AWSClient``, the client will attempt to
    get the key from the ``AWS_ACCESS_KEY_ID`` environment variable. If that is
    not set, it will attempt to get the key from the AWS CLI credentials file.
    The path to the credentials file can be overridden in the
    ``AWS_SHARED_CREDENTIALS_FILE`` environment variable. Note that a value
    set in ``AWS_ACCESS_KEY_ID`` will only be used if there is an accompanying
    value in ``AWS_SECRET_ACCESS_KEY`` environment variable.

    Like the access key, the secret key can be set when creating a new client
    instance. The configuration logic matches the access key with the exception
    of the environment variable. The secret key can set in the
    ``AWS_SECRET_ACCESS_KEY`` environment variable.

    If there is no local configuration or credentials, the client will attempt
    to load the information from the EC2 instance meta-data API, if it is
    available.

    The ``endpoint`` argument is primarily used for testing and allows for
    the use of a specified base URL value instead of the auto-construction of
    a URL using the service and region variables.

    :param str service: The service for the API calls
    :param str profile: Specify the configuration profile name
    :param str region: The AWS region to make requests to
    :param str access_key: The access key
    :param str secret_key: The secret access key
    :param str endpoint: Override the base endpoint URL
    :raises: :exc:`tornado_aws.exceptions.ConfigNotFound`
    :raises: :exc:`tornado_aws.exceptions.ConfigParserError`
    :raises: :exc:`tornado_aws.exceptions.NoCredentialsError`
    :raises: :exc:`tornado_aws.exceptions.NoProfileError`

    """
    ALGORITHM = 'AWS4-HMAC-SHA256'
    ASYNC = False
    CONNECT_TIMEOUT = 10
    REQUEST_TIMEOUT = 30
    SCHEME = 'https'

    def __init__(self, service, profile=None, region=None,
                 access_key=None, secret_key=None, endpoint=None):
        self._client = self._get_client_adapter()
        self._service = service
        self._profile = profile or os.getenv('AWS_DEFAULT_PROFILE', 'default')
        self._region = region or config.get_region(self._profile)
        self._auth_config = config.Authorization(self._profile, access_key,
                                                 secret_key, self._client)
        self._endpoint_url = self._endpoint(endpoint)
        self._host = self._hostname(self._endpoint_url)

    def fetch(self, method, path='/', query_args=None, headers=None, body=b'',
              recursed=False):
        """Executes a request, returning an
        :py:class:`HTTPResponse <tornado.httpclient.HTTPResponse>`.

        If an error occurs during the fetch, we raise an
        :py:class:`HTTPError <tornado.httpclient.HTTPError>` unless the
        ``raise_error`` keyword argument is set to ``False``.

        :param str method: HTTP request method
        :param str path: The request path
        :param dict query_args: Request query arguments
        :param dict headers: Request headers
        :param bytes body: The request body
        :param bool recursed: Internally invoked if it's a recursive fetch
        :rtype: :class:`~tornado.httpclient.HTTPResponse`
        :raises: :class:`~tornado.httpclient.HTTPError`
        :raises: :class:`~tornado_aws.exceptions.NoCredentialsError`
        :raises: :class:`~tornado_aws.exceptions.AWSError`

        """
        if self._auth_config.needs_credentials():
            self._auth_config.refresh()

        request = self._create_request(method, path, query_args, headers, body)

        try:
            result = self._client.fetch(request, raise_error=True)
            return result
        except (OSError, socket.error) as error:
            LOGGER.error('Error making request: %s', error)
            raise exceptions.RequestException(error=error)
        except httpclient.HTTPError as error:
            need_credentials, aws_error = self._process_error(error)
            if need_credentials and not self._auth_config.local_credentials:
                self._auth_config.reset()
                if not recursed:
                    return self.fetch(method, path, query_args,
                                      headers, body, True)
            raise aws_error if aws_error else error

    def close(self):
        """Closes the underlying HTTP client, freeing any resources used."""
        self._client.close()

    def _process_error(self, error):
        """Attempt to process the error coming from AWS. Returns ``True``
        if the client should attempt to fetch credentials and the AWSError
        exception to raise if the client did not have an authentication error.

        :param tornado.httpclient.HTTPError error: The HTTP error
        :rtype: (tuple, tornado_aws.exceptions.AWSError)

        """
        LOGGER.error('Error: %r', error)
        if error.code == 599:
            return False, None
        elif error.code == 400 and self._awz_response(error.response):
            awz_error = self._parse_awz_error(error.response.body)
            return ((awz_error and
                     awz_error['__type'] in _REFRESH_EXCEPTIONS),
                    exceptions.AWSError(
                        type=awz_error['__type'],
                        message=awz_error.get('message', '(null)')))
        try:
            xml_error = self._parse_xml_error(error.response.body)
        except ValueError:
            LOGGER.debug('Could not fallback to XML: %r', error)
            return False, None

        return ((xml_error and
                xml_error['Code'] in _REFRESH_XML_EXCEPTIONS),
                self._aws_error_from_xml(xml_error))

    @staticmethod
    def _aws_error_from_xml(error):
        """Return an AWSError exception for an XML error response, given the
        variation in field names.

        :param dict error: The parsed XML error
        :rtype: tornado_aws.exceptions.AWSError

        """
        return exceptions.AWSError(
            type=error['Code'], message=error['Message'],
            request_id=error.get('RequestId', error.get('x-amzn-RequestId')),
            resource=error.get('Resource'))

    @staticmethod
    def _awz_response(response):
        """Returns ``True`` if the HTTPResponse headers indicate it is
        an AWS style response

        :param tornado.httpclient.HTTPResponse: The HTTP response
        :rtype: bool

        """
        return response.headers.get('Content-Type') in _AWZ_CONTENT_TYPES

    @staticmethod
    def _parse_awz_error(content):
        """Returns the AWZ error parsed out of the HTTPError that was raised.

        :param bytes content: The response error content
        :rtype: dict|None

        """
        payload = json.loads(content.decode('utf-8'))
        if isinstance(payload, dict) and '__type' in payload:
            if '#' in payload['__type']:
                payload['__type'] = \
                    payload['__type'][payload['__type'].index('#') + 1:]
            return payload

    @staticmethod
    def _parse_xml_error(content):
        """Returns the XML error parsed out of the HTTPError that was raised.

        :param bytes content: The response error content
        :rtype: dict
        :raises: ValueError

        """
        payload = txml.loads(content.decode('utf-8'))
        if 'Error' in payload:
            return payload['Error']
        elif 'Errors' in payload and 'Error' in payload['Errors']:
            return payload['Errors']['Error']
        elif 'Response' in payload and 'Errors' in payload['Response'] \
                and 'Error' in payload['Response']['Errors']:
            payload['Response']['Errors']['Error']['RequestId'] = \
                payload['Response'].get('RequestID')
            return payload['Response']['Errors']['Error']
        key = tuple(payload.keys())[0]
        if 'Message' in payload[key]:
            return {'Code': key, 'Message': payload[key]['Message']}
        raise ValueError

    def _auth_header(self, amz_date, date_stamp, request_hash, signed_headers):
        """Return the Authorization string header value

        :param str amz_date: The x-amz-date header value
        :param str date_stamp: The signing date_stamp
        :param str request_hash: The SHA-256 request hash
        :param str signed_headers: A semicolon delimited list of header keys
        :rtype: str

        """
        scope, signature = self._signature(amz_date, date_stamp, request_hash)
        return _HEADER_FORMAT.format(self.ALGORITHM,
                                     self._auth_config.access_key,
                                     scope,
                                     signed_headers, signature)

    def _create_request(self, method, path='/', query_args=None, headers=None,
                        body=b''):
        """Create the HTTPRequest instance that will be used to make the AWS
        API request.

        :param str method: HTTP request method
        :param str path: The request path
        :param dict query_args: Request query arguments
        :param dict headers: Request headers
        :param bytes body: The request body
        :rtype: tornado.httpclient.HTTPClient

        """
        if headers is None:
            headers = {}
        signed_headers, signed_url = self._signed_request(
            method, path, query_args or {}, dict(headers), body or b'')
        LOGGER.debug('Signed URL: %s', signed_url)
        return httpclient.HTTPRequest(signed_url, method,
                                      signed_headers, body,
                                      connect_timeout=self.CONNECT_TIMEOUT,
                                      request_timeout=self.REQUEST_TIMEOUT)

    def _endpoint(self, endpoint):
        """Return the user specified endpoint or dynamically create the
        endpoint from the service and region.

        :rtype: str

        """
        if endpoint:
            return endpoint
        return '{}://{}.{}.amazonaws.com'.format(self.SCHEME,
                                                 self._service,
                                                 self._region)

    def _get_client_adapter(self):
        """Return a HTTP client

        :rtype: :py:class:`tornado.httpclient.HTTPClient`

        """
        return httpclient.HTTPClient(force_instance=True)

    @staticmethod
    def _hostname(url):
        """Parse the url returning a named tuple with the parts of the
        the parsed URL

        :param str url: The URL to parse
        :return: str

        """
        return urlparse(url).netloc

    @staticmethod
    def _quote(value):
        """Return the percent encoded value, ensuring there are no skipped
        characters.

        :param str value: The value to quote
        :rtype: str

        """
        return quote(value, safe='').replace('%7E', '~')

    @staticmethod
    def _sign(key, msg):
        """Sign the msg with the key

        :param bytes key: The signing key
        :param bytes msg: The value to sign
        :return: bytes

        """
        return hmac.new(key, msg, hashlib.sha256).digest()

    def _signed_request(self, method, path, query_args, headers, body):
        """Create the request signature headers and return updated headers
         for the request.

        :param str method: HTTP request method
        :param str path: The request path
        :param dict query_args: Query string args
        :param dict headers: Request headers
        :param bytes body: The request body
        :rtype: dict

        """
        if PYTHON3 and isinstance(body, str):
            body = body.encode('utf-8')

        query_string = self._query_string(query_args)

        timestamp = datetime.datetime.utcnow()
        amz_date = timestamp.strftime('%Y%m%dT%H%M%SZ')
        date_stamp = timestamp.strftime('%Y%m%d')

        payload_hash = hashlib.sha256(body).hexdigest()

        headers.update({
            'Content-Length': str(len(body)),
            'Date': amz_date,
            'X-Amz-Content-sha256': payload_hash
        })

        headers.setdefault('Host', self._host)

        # Temporary auth security token
        if self._auth_config.security_token:
            headers['X-Amz-Security-Token'] = self._auth_config.security_token

        signed_headers, headers_string = self._signed_headers(headers)

        request = '\n'.join([method, path, query_string, headers_string,
                             signed_headers, payload_hash])
        request_hash = hashlib.sha256(request.encode('utf-8')).hexdigest()

        headers['Authorization'] = self._auth_header(amz_date, date_stamp,
                                                     request_hash,
                                                     signed_headers)

        return headers, '{0}{1}?{2}'.format(self._endpoint_url, path,
                                            query_string)

    def _query_string(self, query_args):
        """Return the sorted query string from the query args dict

        :param dict query_args: The dict of query arguments
        :rtype: str

        """
        return '&'.join(['{0}={1}'.format(self._quote(k),
                                          self._quote(query_args[k]))
                         for k in sorted(query_args.keys())])

    def _signature(self, amz_date, date_stamp, request_hash):
        """Return the request scope and signature

        :param str date_stamp: The signing date stamp
        :param str amz_date: The x-amz-date header value
        :param str request_hash: The canonical request signature hash
        :rtype: str, str

        """
        scope = '/'.join([date_stamp, self._region, self._service,
                          'aws4_request'])
        to_sign = '\n'.join([self.ALGORITHM, amz_date, scope, request_hash])
        signing_key = self._signing_key(date_stamp)
        return scope, hmac.new(signing_key, to_sign.encode('utf-8'),
                               hashlib.sha256).hexdigest()

    @staticmethod
    def _signed_headers(headers):
        """Create and return the canonical headers string and the signed
        headers list.

        Canonical header names must be trimmed and lowercase, and sorted in
        ASCII order.

        The signed headers lists the headers in the canonical_headers list,
        delimited with ";" and in alpha order.

        :param dict headers: The request headers
        :rtype: str, str

        """
        tmp = {key.lower(): value for key, value in headers.items()}
        signed_headers = ';'.join([k.lower() for k in sorted(tmp.keys())])
        headers_string = '\n'.join(['{0}:{1}'.format(k, tmp[k])
                                    for k in sorted(tmp.keys())]) + '\n'
        return signed_headers, headers_string

    def _signing_key(self, date_stamp):
        """Create the signature key for the request.

        :param str date_stamp: Date in %Y%m%d format for signing
        :rtype: bytes

        """
        key = 'AWS4{0}'.format(self._auth_config.secret_key)
        date = self._sign(key.encode('utf-8'), date_stamp.encode('utf-8'))
        region = self._sign(date, self._region.encode('utf-8'))
        service = self._sign(region, self._service.encode('utf-8'))
        return self._sign(service, b'aws4_request')


class AsyncAWSClient(AWSClient):
    """Implement a low level AWS client that performs the request signing
    required for AWS API requests.

    ``AWSClient`` uses the same configuration method and environment
    variables as the AWS CLI. For configuration information visit the "Getting
    Set Up" section of the `AWS Command Line Interface user guide
    <http://docs.aws.amazon.com/cli/latest/userguide/>`_.

    When creating the ``AWSClient`` instance you need to specify the
    ``service`` that you will be interacting with. This value is used when
    signing the request headers and must match the service values as specified
    in the `AWS General Reference documentation
    <http://docs.aws.amazon.com/general/latest/gr/Welcome.html>`_.

    The AWS configuration profile can be set when creating the
    ``AWSClient`` instance or by setting the ``AWS_DEFAULT_PROFILE``
    environment variable. If neither are set, ``default`` will be used.

    The AWS access key can be set when creating a new instance. If it's not
    passed in when creating the ``AWSClient``, the client will attempt to
    get the key from the ``AWS_ACCESS_KEY_ID`` environment variable. If that is
    not set, it will attempt to get the key from the AWS CLI credentials file.
    The path to the credentials file can be overridden in the
    ``AWS_SHARED_CREDENTIALS_FILE`` environment variable. Note that a value
    set in ``AWS_ACCESS_KEY_ID`` will only be used if there is an accompanying
    value in ``AWS_SECRET_ACCESS_KEY`` environment variable.

    Like the access key, the secret key can be set when creating a new client
    instance. The configuration logic matches the access key with the exception
    of the environment variable. The secret key can set in the
    ``AWS_SECRET_ACCESS_KEY`` environment variable.

    The ``endpoint`` argument is primarily used for testing and allows for
    the use of a specified base URL value instead of the auto-construction of
    a URL using the service and region variables.

    ``max_clients`` allows for the specification of the maximum number if
    concurrent asynchronous HTTP requests that the client will perform.

    :param str service: The service for the API calls
    :param str profile: Specify the configuration profile name
    :param str region: The AWS region to make requests to
    :param str access_key: The access key
    :param str secret_key: The secret access key
    :param str endpoint: Override the base endpoint URL
    :param int max_clients: Max simultaneous HTTP requests (Default: ``100``)
    :param tornado.ioloop.IOLoop io_loop: Specify the IOLoop to use
    :param bool force_instance: Keep an isolated instance of the HTTP client
    :raises: :exc:`tornado_aws.exceptions.ConfigNotFound`
    :raises: :exc:`tornado_aws.exceptions.ConfigParserError`
    :raises: :exc:`tornado_aws.exceptions.NoCredentialsError`
    :raises: :exc:`tornado_aws.exceptions.NoProfileError`
    :raises: :exc:`tornado_aws.exceptions.CurlNotInstalledError`

    """
    ASYNC = True

    def __init__(self, service, profile=None, region=None, access_key=None,
                 secret_key=None, endpoint=None, max_clients=100,
                 use_curl=False, io_loop=None, force_instance=True):
        self._force_instance = force_instance
        self._ioloop = io_loop or ioloop.IOLoop.current()
        self._max_clients = max_clients
        self._use_curl = use_curl
        if use_curl and not curl_httpclient:
            raise exceptions.CurlNotInstalledError

        super(AsyncAWSClient, self).__init__(
            service, profile, region, access_key, secret_key, endpoint)

    def _get_client_adapter(self):
        """Return an asynchronous HTTP client adapter

        :rtype: :py:class:`tornado.httpclient.AsyncHTTPClient`

        """
        if self._use_curl:
            httpclient.AsyncHTTPClient.configure(
                'tornado.curl_httpclient.CurlAsyncHTTPClient')
        return httpclient.AsyncHTTPClient(
            max_clients=self._max_clients, force_instance=self._force_instance)

    def fetch(self, method, path='/', query_args=None, headers=None, body=None,
              recursed=False):
        """Executes a request, returning an
        :py:class:`HTTPResponse <tornado.httpclient.HTTPResponse>`.

        If an error occurs during the fetch, we raise an
        :py:class:`HTTPError <tornado.httpclient.HTTPError>` unless the
        ``raise_error`` keyword argument is set to ``False``.

        :param str method: HTTP request method
        :param str path: The request path
        :param dict query_args: Request query arguments
        :param dict headers: Request headers
        :param bytes body: The request body
        :param bool recursed: Internal use only
        :rtype: :class:`~tornado.httpclient.HTTPResponse`
        :raises: :class:`~tornado.httpclient.HTTPError`
        :raises: :class:`~tornado_aws.exceptions.AWSError`
        :raises: :class:`~tornado_aws.exceptions.NoCredentialsError`

        """
        future = concurrent.Future()

        def on_response(response):
            exc = response.exception()
            if exc:
                if isinstance(exc, httpclient.HTTPError):
                    need_credentials, aws_error = self._process_error(exc)
                    if need_credentials and \
                            not self._auth_config.local_credentials:
                        self._auth_config.reset()
                        if not recursed:
                            def on_retry(retry):
                                if not self._future_exception(retry, future):
                                    future.set_result(retry.result())

                            request = self.fetch(method, path, query_args,
                                                 headers, body, True)
                            self._ioloop.add_future(request, on_retry)
                            return
                    future.set_exception(aws_error if aws_error else exc)
                else:
                    LOGGER.error('Error making request: %s', exc)
                    future.set_exception(
                        exceptions.RequestException(error=exc))
            else:
                future.set_result(response.result())

        def perform_request():
            request = self._create_request(method, path, query_args, headers,
                                           body)
            api_future = self._client.fetch(request, raise_error=True)
            self._ioloop.add_future(api_future, on_response)

        def on_refreshed(response):
            if not self._future_exception(response, future):
                perform_request()

        if self._auth_config.needs_credentials():
            request_future = self._auth_config.refresh()
            self._ioloop.add_future(request_future, on_refreshed)
        else:
            perform_request()

        return future

    @staticmethod
    def _future_exception(inner, outer):
        exception = inner.exception()
        if exception:
            outer.set_exception(exception)
        return bool(exception)



"""
XML Deserialization
===================

Parse XML return content and return it as a dict.

"""
import collections
import xml.etree.ElementTree as ET


def loads(content):
    """Return the XML document returned from AWS as a dict

    :param str content: Response content from AWS
    :rtype: dict
    :raises: ValueError

    """
    try:
        return _xml_to_dict(ET.XML(content))
    except ET.ParseError as error:
        raise ValueError(str(error))


def _xml_to_dict(t):
    """Process child nodes, initially taken from
    https://stackoverflow.com/a/10076823/13203

    :param xml.etree.ElementTree.Element t: The XML node to process

    """
    d = {t.tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = collections.defaultdict(list)
        for dc in map(_xml_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if t.attrib:
        d[t.tag].update(('@' + k, v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if (children or t.attrib) and text:
            d[t.tag]['#text'] = text
        else:
            d[t.tag] = t.text.strip()
    return dict(d)



"""
AWS Credentials Loader

"""
try:
    import configparser
except ImportError:  # pragma: no cover
    import ConfigParser as configparser
import json
import logging
import os
from os import path
import socket

from tornado import concurrent, httpclient, ioloop
try:
    from tornado import curl_httpclient
except ImportError:  # pragma: no cover
    curl_httpclient = httpclient
    curl_httpclient.CurlAsyncHTTPClient = httpclient.AsyncHTTPClient

from tornado_aws import exceptions

LOGGER = logging.getLogger(__name__)

DEFAULT_CREDENTIALS_PATH = '~/.aws/credentials'
DEFAULT_REGION = 'us-east-1'
INSTANCE_ENDPOINT = 'http://169.254.169.254/latest/{}'
INSTANCE_ROLE_PATH = '/meta-data/iam/security-credentials/'
INSTANCE_CREDENTIALS_PATH = '/meta-data/iam/security-credentials/{}'
REGION_PATH = '/dynamic/instance-identity/document'

HTTP_TIMEOUT = 0.25


def get_region(profile):
    """Return the credentials from the configured ~/.aws/credentials file
    following a similar behavior implemented by awscli and botocore.

    :param str profile: Use the optional profile for getting settings
    :return: region
    :rtype: str
    :raises: exceptions.ConfigNotFound

    """
    region = os.getenv('AWS_DEFAULT_REGION', None)
    if region:
        return region

    file_path = os.getenv('AWS_CONFIG_FILE', '~/.aws/config')
    try:
        config = _parse_file(file_path)
    except exceptions.ConfigNotFound:
        try:
            return _request_region_from_instance()
        except (httpclient.HTTPError, OSError, socket.error) as error:
            LOGGER.error('Error fetching from EC2 Instance Metadata (%s)',
                         error)
            raise exceptions.ConfigNotFound(path=file_path)

    key = 'profile {0}'.format(profile)
    if key not in config and 'default' not in config:
        raise exceptions.NoProfileError(path=file_path, profile=profile)
    return config.get(
        key, {}).get('region') or config.get(
            'default', {}).get('region') or DEFAULT_REGION


def _is_async_client(client):
    """Returns ``True`` if the client that is passed in is asynchronous.

    :param client: The HTTP client to use for EC2 API
    :type client: tornado.httpclient.HTTPClient or
        tornado.httpclient.AsyncHTTPClient or
        tornado.curl_httpclient.CurlAsyncHTTPClient
    :rtype: bool

    """
    return isinstance(client, (httpclient.AsyncHTTPClient,
                               curl_httpclient.CurlAsyncHTTPClient))


def _parse_file(file_path):
    """Parse the specified configuration file, returning a nested dict
    of key/value pairs by section.

    :param str file_path: The path of the file to read.
    :rtype: dict

    """
    file_path = path.abspath(path.expanduser(path.expandvars(file_path)))
    LOGGER.debug('Attempting to load credentials from %s', file_path)
    if not path.exists(file_path):
        raise exceptions.ConfigNotFound(path=file_path)

    parser = configparser.RawConfigParser()
    try:
        parser.read(file_path)
    except configparser.Error as error:
        LOGGER.error('Error reading file: %s', error)
        raise exceptions.ConfigParserError(path=file_path)

    config = {}
    for section in parser.sections():
        config[section] = {}
        for option in parser.options(section):
            config[section][option] = parser.get(section, option)
    return config


def _request_region_from_instance():
    """Attempt to get the region from the instance metadata

    :rtype: str

    """
    url = INSTANCE_ENDPOINT.format(REGION_PATH)
    response = httpclient.HTTPClient().fetch(
        url, connect_timeout=HTTP_TIMEOUT, request_timeout=HTTP_TIMEOUT)
    data = json.loads(response.body.decode('utf-8'))
    return data['region']


class Authorization(object):
    """Object used to hold configuration information."""

    def __init__(self, profile, access_key=None, secret_key=None, client=None):
        """Create a new instance of the ``_AuthConfig`` class.

        :param str profile: The configuration profile to use
        :param str access_key: Optional configured access key
        :param str secret_key: Optional configured secret key
        :param client: The HTTP client to use for EC2 API
        :type client: tornado.httpclient.HTTPClient or
            tornado.httpclient.AsyncHTTPClient

        """
        self._client = client
        self._profile = profile
        self._config_credentials = False
        self._access_key, self._secret_key = self._resolve_credentials(
            access_key, secret_key)
        self._local_credentials = bool(self._access_key) and \
            not self._config_credentials
        self._security_token = None
        self._expiration = None
        self._is_async = _is_async_client(client)
        self._ioloop = ioloop.IOLoop.current() if self._is_async else None

    @property
    def access_key(self):
        """Return the current access key.

        :rtype: str or None

        """
        return self._access_key

    @property
    def local_credentials(self):
        """Indicates if the credentials are loaded dynamically or if they
        are statically set upon initialization

        :rtype: bool

        """
        return self._local_credentials

    @property
    def secret_key(self):
        """Return the current secret key.

        :rtype: str or None

        """
        return self._secret_key

    @property
    def security_token(self):
        """Return the current security token value.

        :rtype: str or None

        """
        return self._security_token

    def needs_credentials(self):
        """Returns True if the client needs to fetch remote credentials.

        :rtype: bool

        """
        if self._local_credentials:
            return False
        return not self._access_key

    def refresh(self):
        """Load dynamic credentials from the AWS Instance Metadata and user
        data HTTP API.

        :raises: tornado_aws.exceptions.NoCredentialsError

        """
        future = concurrent.Future() if self._is_async else None

        # Refresh config file credentials
        if self._config_credentials:
            self._access_key, self._secret_key = \
                self._resolve_config_file_credentials()
            future.set_result(True)
            return future

        LOGGER.debug('Refreshing EC2 IAM Credentials')
        try:
            result = self._fetch_credentials()
            if concurrent.is_future(result):

                def on_complete(response):
                    exception = response.exception()
                    if exception:
                        if isinstance(exception, httpclient.HTTPError) and \
                                exception.code == 599:
                            future.set_exception(
                                exceptions.NoCredentialsError())
                        else:
                            future.set_exception(exception)
                        return
                    self._assign_credentials(response.result())
                    future.set_result(True)

                self._ioloop.add_future(result, on_complete)
            else:
                self._assign_credentials(result)
        except (httpclient.HTTPError,
                OSError) as error:
            LOGGER.error('Error Fetching Credentials: %s', error)
            raise exceptions.NoCredentialsError
        return future

    def reset(self):
        """Reset the security credentials.

        :raises: tornado_aws.exceptions.LocalCredentialsError()

        """
        if self.local_credentials:
            raise exceptions.LocalCredentialsError
        self._access_key = None
        self._secret_key = None
        self._expiration = None
        self._security_token = None

    def _assign_credentials(self, data):
        """Assign the values returned by the EC2 Metadata and user data API to
        the internal credentials attributes.

        :param dict data: The data returned by the EC2 api

        """
        self._access_key = data['AccessKeyId']
        self._secret_key = data['SecretAccessKey']
        self._expiration = data['Expiration']
        self._security_token = data['Token']

    def _fetch_credentials(self):
        """Fetch credential information from the local EC2 Metadata and user
        data API.

        :rtype: dict

        """
        if self._is_async:
            return self._fetch_credentials_async()
        role = self._get_role()
        credentials = self._get_instance_credentials(role)
        return credentials

    def _fetch_credentials_async(self):
        """Return the credentials from the EC2 Instance Metadata and user data
        API using an Async adapter.

        :return: :class:`~concurrent.Future`

        """
        future = concurrent.Future()

        def on_credentials(response):
            if not self._future_exception(response, future):
                result = response.result()
                future.set_result(result)

        def on_role(response):
            if not self._future_exception(response, future):
                req = self._get_instance_credentials_async(response.result())
                self._ioloop.add_future(req, on_credentials)

        request = self._get_role_async()
        self._ioloop.add_future(request, on_role)
        return future

    @staticmethod
    def _future_exception(inner, outer):
        exception = inner.exception()
        if exception:
            outer.set_exception(exception)
        return bool(exception)

    def _get_instance_credentials(self, role):
        """Attempt to get temporary credentials for the specified role from the
        EC2 Instance Metadata and user data API

        :param str role: The role to get temporary credentials for

        :rtype: dict
        :raises: tornado.httpclient.HTTPError

        """
        url_path = INSTANCE_CREDENTIALS_PATH.format(role)
        response = self._client.fetch(INSTANCE_ENDPOINT.format(url_path),
                                      connect_timeout=HTTP_TIMEOUT,
                                      request_timeout=HTTP_TIMEOUT)
        return json.loads(response.body.decode('utf-8'))

    def _get_instance_credentials_async(self, role):
        """Attempt to get temporary credentials for the specified role from the
        EC2 Instance Metadata and user data API

        :param str role: The role to get temporary credentials for

        :rtype: :class:`~tornado.concurrent.Future`
        :raises: tornado.httpclient.HTTPError

        """
        future = concurrent.Future()

        def on_response(response):
            if not self._future_exception(response, future):
                body = response.result().body
                future.set_result(json.loads(body.decode('utf-8')))

        url_path = INSTANCE_CREDENTIALS_PATH.format(role)
        result = self._client.fetch(INSTANCE_ENDPOINT.format(url_path),
                                    connect_timeout=HTTP_TIMEOUT,
                                    request_timeout=HTTP_TIMEOUT)
        self._ioloop.add_future(result, on_response)
        return future

    def _get_role(self):
        """Fetch the IAM role from the ECS Metadata and user data API

        :param tornado.httpclient.HTTPClient client: The HTTP client
        :rtype: str
        :raises: tornado.httpclient.HTTPError

        """
        url = INSTANCE_ENDPOINT.format(INSTANCE_ROLE_PATH)
        response = self._client.fetch(url,
                                      connect_timeout=HTTP_TIMEOUT,
                                      request_timeout=HTTP_TIMEOUT)
        return response.body.decode('utf-8')

    def _get_role_async(self):
        """Fetch the IAM role from the ECS Metadata and user data API

        :rtype: :class:`~tornado.concurrent.Future`
        :raises: tornado.httpclient.HTTPError

        """
        future = concurrent.Future()

        def on_response(response):
            if not self._future_exception(response, future):
                role = response.result()
                future.set_result(role.body.decode('utf-8'))

        url = INSTANCE_ENDPOINT.format(INSTANCE_ROLE_PATH)
        request = self._client.fetch(url,
                                     connect_timeout=HTTP_TIMEOUT,
                                     request_timeout=HTTP_TIMEOUT)
        self._ioloop.add_future(request, on_response)
        return future

    def _resolve_credentials(self, access_key, secret_key):
        """Try and load the credentials file from disk checking first to see
        if a path is specified in the ``AWS_SHARED_CREDENTIALS_FILE``
        environment variable and if not, falling back to ``~/.aws/credentials``

        :return: access_key, secret_key
        :rtype: str, str

        """
        access_key = os.getenv('AWS_ACCESS_KEY_ID', access_key)
        secret_key = os.getenv('AWS_SECRET_ACCESS_KEY', secret_key)
        if access_key and secret_key:
            return access_key, secret_key
        return self._resolve_config_file_credentials()

    def _resolve_config_file_credentials(self):
        """Attempt to load credentials via configuration file, looking to
        the AWS_SHARED_CREDENTIALS_FILE environment variable for the path
        or defaulting to `'~/.aws/credentials`. This allows for instances
        where the credentials file is mounted or managed by an external
        system and can change while the application is running.

        :raises: ConfigNotFound
        :raises: ConfigParserError
        :return: access_key, secret_key
        :rtype: str, str

        """
        file_path = os.getenv('AWS_SHARED_CREDENTIALS_FILE',
                              DEFAULT_CREDENTIALS_PATH)
        try:
            config = _parse_file(file_path)
        except exceptions.ConfigNotFound:
            return None, None

        if self._profile not in config:
            raise exceptions.NoProfileError(path=file_path,
                                            profile=self._profile)

        values = []
        for key in ['aws_access_key_id', 'aws_secret_access_key']:
            values.append(config[self._profile].get(key) or
                          config.get('default', {}).get(key))
        if not all(values):
            return None, None
        self._config_credentials = True
        return values[0], values[1]



from os import path

import setuptools

CLASSIFIERS = ['Development Status :: 5 - Production/Stable',
               'Intended Audience :: Developers',
               'License :: OSI Approved :: BSD License',
               'Operating System :: OS Independent',
               'Programming Language :: Python :: 2',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3',
               'Programming Language :: Python :: 3.5',
               'Programming Language :: Python :: 3.6',
               'Programming Language :: Python :: 3.7',
               'Programming Language :: Python :: Implementation :: CPython',
               'Programming Language :: Python :: Implementation :: PyPy',
               'Topic :: Communications',
               'Topic :: Internet',
               'Topic :: Software Development :: Libraries']

DESC = 'A low-level Amazon Web Services API client for Tornado'


def read_requirements(name):
    requirements = []
    try:
        with open(path.join('requires', name)) as req_file:
            for line in req_file:
                if '#' in line:
                    line = line[:line.index('#')]
                line = line.strip()
                if line.startswith('-r'):
                    requirements.extend(read_requirements(line[2:].strip()))
                elif line and not line.startswith('-'):
                    requirements.append(line)
    except IOError:
        pass
    return requirements


setuptools.setup(
    name='tornado-aws',
    version='1.4.0',
    description=DESC,
    long_description=open('README.rst').read(),
    author='Gavin M. Roy',
    author_email='gavinmroy@gmail.com',
    url='https://tornado-aws.readthedocs.io',
    packages=['tornado_aws'],
    package_data={'': ['LICENSE', 'README.rst', 'requires/installation.txt']},
    include_package_data=True,
    install_requires=read_requirements('requires/installation.txt'),
    extras_require={'curl': ['pycurl']},
    tests_require=read_requirements('requires/testing.txt'),
    license='BSD',
    classifiers=CLASSIFIERS,
    zip_safe=True)
